/// <reference types="node" />
import EventSource from '@frando/eventsource';
import { EventEmitter } from 'events';
import { FetchOpts } from './fetch.js';
import { Collection } from './collection.js';
import { Bots } from './bots.js';
import type { Logger } from '@arsonar/common';
export interface CreateCollectionOpts {
    key?: string;
    name?: string;
    alias?: string;
}
export interface CollectionConfig {
    share: boolean;
}
export interface OpenCollectionOpts {
    reset?: boolean;
}
export interface EventSourceOpts {
    endpoint?: string;
    token?: string;
    onmessage?: (event: any) => void;
    onerror?: (error: any) => void;
    withCredentials?: boolean | undefined;
    headers?: object | undefined;
    proxy?: string | undefined;
    https?: object | undefined;
    rejectUnauthorized?: boolean | undefined;
}
export interface WorkspaceOpts {
    workspace?: string;
    endpoint?: string;
    url?: string;
    token?: string;
    accessCode?: string;
    id?: string;
    log?: Logger;
}
export declare class Workspace extends EventEmitter {
    endpoint: string;
    log: Logger;
    bots: Bots;
    opened: boolean;
    private _collections;
    private _token?;
    private _accessCode?;
    private _eventSources;
    private _openPromise?;
    /**
       * A Sonar workspace. Provides methods to open collection under this endpoint.
       *
       * @constructor
       * @param {object} [opts] - Optional options.
       * @param {string} [opts.url=http://localhost:9191/api/v1/default] - The API endpoint to talk to.
       * @param {string} [opts.accessCode] - An access code to login at the endpoint.
       * @param {string} [opts.token] - A JSON web token to authorize to the endpoint.
       * @param {string} [opts.name] - The name of this client.
       */
    constructor(opts?: WorkspaceOpts);
    get token(): string | undefined;
    /**
       * Closes the client.
       */
    close(): Promise<void>;
    /**
       * Closes the client.
       */
    open(): Promise<void>;
    private _open;
    private _login;
    /**
       * Get a list of all collections available on this endpoint.
       *
       * @async
       * @return {Promise.<object[]>} Promise that resolves to an array of collection info objects.
       */
    listCollections(): Promise<any>;
    /**
       * Creates a collection with name name on the Sonar server. The name may not contain whitespaces. opts is an optional object with:
       *
       * @async
       * @param {string} name - Name of the new collection, may not contain whitespaces.
       * @param {object} [opts] - Optional options object.
       * @param {string} [opts.key] - Hex string of an existing collection. Will then sync this collection instead of creating a new, empty collection.
       * @param {string} [opts.alias] - When setting key, alias is required and is your nick name within this collection.
       * @return {Promise<Collection>} The created collection.
       */
    createCollection(name: string, opts?: CreateCollectionOpts): Promise<Collection>;
    /**
       * Updates the config of a collection.
       *
       * @async
       * @param {string} name - Name of the collection.
       * @param {object} info - [TODO:description]
       * @param {boolean} info.share - Controls whether a collection is shared via p2p.
       * @return {Promise<void>}
       */
    updateCollection(name: string, config: CollectionConfig): Promise<CollectionConfig>;
    /**
       * Returns a Collection object for a given key or name of a collection.
       *
       * @async
       * @param {string} keyOrName - Key or name of the collection to open/return.
       * @return {Promise<Collection>}
       */
    openCollection(keyOrName: string, opts?: OpenCollectionOpts): Promise<Collection>;
    getCollection(keyOrName: string): Collection | undefined;
    getAuthHeaders(opts?: Record<string, string>): Record<string, string>;
    getHeaders(): Record<string, string>;
    /**
       * Fetch a resource.
       *
       * This is a wrapper around the fetch web API. It should be API compatible to fetch,
       * with the following changes:
       *
       * @async
       * @param {string} [opts.requestType='json'] Request encoding and content type.
       *   Supported values are 'json' and 'binary'
       * @param {string} [opts.responseType='text'] Response encoding. If the response
       *    has a JSON content type, will always be set to 'json'.
       *    Supported values are 'text', 'binary' and 'stream'.
       * @param {object} [opts.params] Query string parameters (will be encoded correctly).
       *
       * @return {Promise<object>} If the response has a JSON content type header, the
       *    decoded JSON will be returned. if opts.responseType is 'binary' or 'text',
       *    the response will be returned as a buffer or text.
       *
       * TODO: Rethink the default responseType cascade.
       */
    fetch(url: string, opts?: FetchOpts): Promise<any>;
    createEventSource(path: String, opts?: EventSourceOpts): EventSource;
}
//# sourceMappingURL=workspace.d.ts.map