/// <reference types="node" />
import type { Collection } from './collection';
import type { Readable } from 'streamx';
import { FetchOpts } from './fetch';
export interface UploadFileOpts {
    onUploadProgress?: (progress: any) => void;
}
export interface RangeOpts {
    range?: {
        from: number;
        to: number;
    };
}
export declare type FileMetadata = any;
export declare type FileBody = string | Readable | Uint8Array | Buffer | ReadableStream | any;
/**
 * File system for a collection.
 */
export declare class Files {
    collection: Collection;
    /**
       * File system for a collection.
       *
       * @constructor
       * @param {Collection} collection - Collection
       */
    constructor(collection: Collection);
    /**
       * Create a new file
       * @param {FileBody|Buffer} stream - File content as stream or buffer
       * @param {object} [metadata] - File record metadata (see file record schema)
       * @param {object} [opts] - Options.
       *  - onUploadProgress: Callback to invoke with upload progress information
       * @returns {Record} - The created file record
       */
    createFile(stream: FileBody, metadata?: any, opts?: UploadFileOpts): Promise<any>;
    /**
       * Update a file
       * @param {string} id - The file record id
       * @param {FileBody|Buffer} stream - File content as stream or buffer
       * @param {object} [metadata] - File record metadata (see file record schema)
       * @param {object} [opts] - Options.
       *  - onUploadProgress: Callback to invoke with upload progress information
       * @returns {Record} - The created file record
       */
    updateFile(id: string, stream: FileBody, metadata: any, opts?: UploadFileOpts): Promise<any>;
    /**
       * Read a file into a buffer.
       *
       * @async
       * @param {string} id - A file ID
       * @param {object} [opts] - Options. TODO: document.
       * @throws Will throw if the path is not found.
       * @return {Promise<ArrayBuffer|Buffer>} The file content. A Buffer object in Node.js, a ArrayBuffer object in the browser.
       */
    readFile(id: string): Promise<Readable<Uint8Array>>;
    readFile(id: string, opts?: RangeOpts & FetchOpts & {
        responseType: undefined;
    }): Promise<Readable<Uint8Array>>;
    readFile(id: string, opts?: RangeOpts & FetchOpts & {
        responseType: 'text';
    }): Promise<string>;
    readFile(id: string, opts?: RangeOpts & FetchOpts & {
        responseType: 'buffer';
    }): Promise<Uint8Array>;
    readFile(id: string, opts?: RangeOpts & FetchOpts & {
        responseType: 'stream';
    }): Promise<ReadableStream<Uint8Array>>;
    readFile(id: string, opts?: RangeOpts & FetchOpts & {
        responseType: 'raw';
    }): Promise<Response>;
    /**
       * Returns the HTTP url for a file.
       * @async
       * @param {string} id - A file ID
       * @return {Promise<string>} The file URL
       */
    getURL(id: string, includeToken?: boolean): string;
    /**
       * Get the metadata for a file
       *
       * @async
       * @param {string} id - A file ID
       * @throws Will throw if the path is not found.
       * @return {Promise<object>} The file record value (metadata)
       */
    getFileMetadata(id: string, opts?: FetchOpts): Promise<FileMetadata>;
}
//# sourceMappingURL=files.d.ts.map