/**
 * File system for a collection.
 */
export class Files {
    /**
       * File system for a collection.
       *
       * @constructor
       * @param {Collection} collection - Collection
       */
    constructor(collection) {
        this.collection = collection;
        this.collection = collection;
    }
    /**
       * Create a new file
       * @param {FileBody|Buffer} stream - File content as stream or buffer
       * @param {object} [metadata] - File record metadata (see file record schema)
       * @param {object} [opts] - Options.
       *  - onUploadProgress: Callback to invoke with upload progress information
       * @returns {Record} - The created file record
       */
    async createFile(stream, metadata, opts = {}) {
        const requestType = 'stream';
        const params = {};
        if (metadata) {
            params.metadata = JSON.stringify(metadata);
        }
        return await this.collection.fetch('/file', {
            method: 'POST',
            body: stream,
            params,
            requestType
            // onUploadProgress: opts.onUploadProgress
        });
    }
    /**
       * Update a file
       * @param {string} id - The file record id
       * @param {FileBody|Buffer} stream - File content as stream or buffer
       * @param {object} [metadata] - File record metadata (see file record schema)
       * @param {object} [opts] - Options.
       *  - onUploadProgress: Callback to invoke with upload progress information
       * @returns {Record} - The created file record
       */
    async updateFile(id, stream, metadata, opts = {}) {
        const requestType = 'stream';
        const params = {};
        if (metadata) {
            params.metadata = JSON.stringify(metadata);
        }
        return await this.collection.fetch('/file/' + id, {
            method: 'PUT',
            body: stream,
            params,
            requestType
            // onUploadProgress: opts.onUploadProgress
        });
    }
    async readFile(id, opts) {
        opts = opts || {};
        opts.responseType = opts.responseType || 'stream';
        if (!opts.headers) {
            opts.headers = {};
        }
        if (opts.range) {
            opts.headers.Range = `bytes=${opts.range.from || 0}-${opts.range.to || ''}`;
        }
        return await this.collection.fetch('/file/' + id, opts);
    }
    /**
       * Returns the HTTP url for a file.
       * @async
       * @param {string} id - A file ID
       * @return {Promise<string>} The file URL
       */
    getURL(id, includeToken) {
        let url = `${this.collection.endpoint}/file/${id}`;
        if (includeToken && this.collection.workspace.token) {
            url += `?token=${this.collection.workspace.token}`;
        }
        return url;
    }
    /**
       * Get the metadata for a file
       *
       * @async
       * @param {string} id - A file ID
       * @throws Will throw if the path is not found.
       * @return {Promise<object>} The file record value (metadata)
       */
    async getFileMetadata(id, opts = {}) {
        return await this.collection.fetch(`/file/${id}?meta=1`, opts);
    }
}
//# sourceMappingURL=files.js.map